use std::{
    env,
    error::Error,
    fmt::Display,
    fs::{read_dir, read_to_string, File},
    io::Read,
};

use sha2::{Digest, Sha256};

type Result<T, E = Box<dyn Error>> = std::result::Result<T, E>;

/// Usage: keyfork-mnemonic-generate [bitsize: 128 or 256]
/// CHECKS:
/// * If the system is online
/// * If a kernel is running post-BLAKE2
///
/// TODO:
/// * --features kitchen-sink: load username; system time's most random, precise bits; hostname;
///   kernel version; other env specific shit into a CSPRNG

static WARNING_LINKS: [&str; 1] =
    ["https://lore.kernel.org/lkml/20211223141113.1240679-2-Jason@zx2c4.com/"];

#[derive(Debug)]
enum MnemonicGenerationError {
    InvalidByteCount(usize),
    InvalidByteLength(usize),
}

impl MnemonicGenerationError {
    fn boxed(self) -> Box<Self> {
        Box::new(self)
    }
}

impl Display for MnemonicGenerationError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            MnemonicGenerationError::InvalidByteCount(count) => {
                write!(f, "Invalid byte count: {count}, must be divisible by 8")
            }
            MnemonicGenerationError::InvalidByteLength(count) => {
                write!(f, "Invalid byte length: {count}, must be 128 or 256")
            }
        }
    }
}

impl Error for MnemonicGenerationError {}

fn ensure_safe_kernel_version() {
    let kernel_version = read_to_string("/proc/version").expect("/proc/version");
    let v = kernel_version
        .split(' ')
        .nth(2)
        .expect("Unable to parse kernel version")
        .split('.')
        .take(2)
        .map(str::parse)
        .map(|x| x.expect("Unable to parse kernel version number"))
        .collect::<Vec<u32>>();
    let [major, minor, ..] = v.as_slice() else {
        panic!("Unable to determine major and minor: {kernel_version}");
    };
    assert!(
        [major, minor] > [&5, &4],
        "kernel can't generate clean entropy: {}",
        WARNING_LINKS[0]
    );
}

fn ensure_offline() {
    let paths = read_dir("/sys/class/net").expect("Unable to read network interfaces");
    for entry in paths {
        let mut path = entry.expect("Unable to read directory entry").path();
        if path
            .as_os_str()
            .to_str()
            .expect("Unable to decode UTF-8 filepath")
            .split('/')
            .last()
            .unwrap()
            == "lo"
        {
            continue;
        }
        path.push("operstate");
        let isup = read_to_string(&path).expect("Unable to read operstate of network interfaces");
        assert_ne!(isup.trim(), "up", "No network interfaces should be up");
    }
}

// TODO: Can a Mnemonic be formatted using a wordlist, without allocating or without storing the
// entire word list?
//
// NOTE: Yes, use a Language and pass a reference to this.
struct Mnemonic {
    pub words: Vec<usize>,
    wordlist: Vec<String>,
}

impl Mnemonic {
    fn from_entropy(bytes: &[u8]) -> Result<Mnemonic> {
        let bit_count = bytes.len() * 8;
        let hash = generate_slice_hash(bytes);

        if bit_count % 32 != 0 {
            return Err(MnemonicGenerationError::InvalidByteCount(bit_count).boxed());
        }
        // 192 supported for test suite
        if ![128, 192, 256].contains(&bit_count) {
            return Err(MnemonicGenerationError::InvalidByteLength(bit_count).boxed());
        }
        assert_eq!(bit_count % 32, 0, "bit count must be in 32 bit increments");

        let mut bits = vec![false; bit_count + bit_count / 32];

        for byte_index in 0..bit_count / 8 {
            for bit_index in 0..8 {
                bits[byte_index * 8 + bit_index] = (bytes[byte_index] & (1 << (7 - bit_index))) > 0;
            }
        }
        for check_bit in 0..bit_count / 32 {
            bits[bit_count + check_bit] = (hash[check_bit / 8] & (1 << (7 - (check_bit % 8)))) > 0;
        }

        assert_eq!(bits.len() % 11, 0, "unstable bit count");

        let words = bits
            .chunks_exact(11)
            .map(|chunk| bitslice_to_usize(chunk.try_into().expect("11 bit chunks")))
            .collect::<Vec<_>>();

        let wordlist = build_wordlist();
        Ok(Mnemonic { words, wordlist })
    }
}

impl Display for Mnemonic {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut iter = self.words.iter().peekable();
        while let Some(word_index) = iter.next() {
            let word = &self.wordlist[*word_index];
            write!(f, "{word}")?;
            if iter.peek().is_some() {
                write!(f, " ")?;
            }
        }
        Ok(())
    }
}

fn generate_slice_hash(data: &[u8]) -> Vec<u8> {
    let mut hasher = Sha256::new();
    hasher.update(data);
    hasher.finalize().to_vec()
}

fn build_wordlist() -> Vec<String> {
    let wordlist_file = include_str!("wordlist.txt");
    wordlist_file
        .lines()
        .skip(1)
        .map(|x| x.trim().to_string())
        .collect()
}

fn bitslice_to_usize(bitslice: [bool; 11]) -> usize {
    let mut index = 0usize;
    for i in 0..11 {
        index += usize::from(bitslice[10 - i]) << i;
    }
    index
}

fn main() -> Result<()> {
    if !env::vars()
        .any(|(name, _)| name == "SHOOT_SELF_IN_FOOT" || name == "INSECURE_HARDWARE_ALLOWED")
    {
        ensure_safe_kernel_version();
        ensure_offline();
    }

    let bit_size: usize = env::args()
        .nth(1)
        .unwrap_or(String::from("256"))
        .parse()
        .expect("Expected integer bit size");
    assert!(
        bit_size == 128 || bit_size == 256,
        "Only 12 or 24 word mnemonics are supported"
    );

    let mut random_handle = File::open("/dev/urandom")?;
    let entropy = &mut [0u8; 256 / 8];
    random_handle.read_exact(&mut entropy[..])?;

    let mnemonic = Mnemonic::from_entropy(&entropy[..bit_size / 8])?;

    println!("{mnemonic}");

    Ok(())
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn has_good_wordlist() {
        let wordlist = build_wordlist();
        assert_eq!(
            wordlist.len(),
            2usize.pow(11),
            "Wordlist did not include correct word count"
        );
    }

    #[test]
    fn loads_mnemonics() -> Result<()> {
        let content = include_str!("test/vectors.json");
        let jsonobj: serde_json::Value = serde_json::from_str(content)?;

        for test in jsonobj["english"].as_array().unwrap() {
            let [ref hex_, ref seed, ..] = test.as_array().unwrap()[..] else {
                panic!("bad test: {test}");
            };
            let hex = hex::decode(hex_.as_str().unwrap()).unwrap();

            let mnemonic = Mnemonic::from_entropy(&hex)?;

            assert_eq!(mnemonic.to_string(), seed.as_str().unwrap());
        }
        Ok(())
    }

    #[test]
    fn matches_bip39_crate() {
        let mut random_handle = File::open("/dev/random").unwrap();
        let entropy = &mut [0u8; 256 / 8];
        random_handle.read_exact(&mut entropy[..]).unwrap();
        let my_mnemonic = super::Mnemonic::from_entropy(&entropy[..256 / 8]).unwrap();
        let their_mnemonic = bip39::Mnemonic::from_entropy(&entropy[..256 / 8]).unwrap();
        /*
        let my_words = my_mnemonic.words.clone();
        let their_words = their_mnemonic.word_iter().collect::<Vec<_>>();
        */
        assert_eq!(my_mnemonic.to_string(), their_mnemonic.to_string());
    }

    #[test]
    fn count_to_get_duplicate_words() {
        let mut count = 0.;
        let tests = 10_000;
        let mut random_handle = File::open("/dev/random").unwrap();
        let entropy = &mut [0u8; 256 / 8];
        for _ in 0..tests {
            random_handle.read_exact(&mut entropy[..]).unwrap();
            // let bits = generate_slice_hash(entropy);
            let mnemonic = Mnemonic::from_entropy(&entropy[..256 / 8]).unwrap();
            let mut words = mnemonic.words.clone();
            words.sort();
            assert_eq!(words.len(), 24);
            words.dedup();
            if words.len() != 24 {
                count += 1.;
            }
        }
        // NOTE: Birthday problem math is: 0.126532
        // Set values to (about) 1 below, 1 above
        // Source: https://en.wikipedia.org/wiki/Birthday_problem
        let min = 11.5;
        let max = 13.5;
        assert!(
            count > f64::from(tests) * min / 100.,
            "{count} probability should be more than {min}%: {}",
            count / f64::from(tests)
        );
        assert!(
            count < f64::from(tests) * max / 100.,
            "{count} probability should be more than {max}%: {}",
            count / f64::from(tests)
        );
    }
}
